# pylint: disable=missing-class-docstring, missing-function-docstring, missing-module-docstring
# pylint: disable=invalid-name, line-too-long

import unittest

from dexipy.values import DexiValues, DexiValueType
from dexipy.values import dexi_value_type, check_dexi_value, check_dexi_value_specification
from dexipy.values import dexi_value_as_set, dexi_value_as_distr, reduce_dexi_value, aggregate_value, reduce_set
from dexipy.values import compare_values, compare_values_by_preference
from dexipy.types import DexiOrder

class test_test_values(unittest.TestCase):

    def test_dexi_value_type(self):
        self.assertEqual(dexi_value_type(None), DexiValueType.NONE)
        self.assertEqual(dexi_value_type(5), DexiValueType.INT)
        self.assertEqual(dexi_value_type(5.5), DexiValueType.FLOAT)
        self.assertEqual(dexi_value_type(set()), DexiValueType.SET)
        self.assertEqual(dexi_value_type({1,2}), DexiValueType.SET)
        self.assertEqual(dexi_value_type([1,2]), DexiValueType.LIST)
        self.assertEqual(dexi_value_type([]), DexiValueType.LIST)
        self.assertEqual(dexi_value_type({1: 2}), DexiValueType.DICT)
        self.assertEqual(dexi_value_type(lambda x: x), DexiValueType.ERROR)

    def test_check_dexi_value(self):
        self.assertTrue(check_dexi_value(None))
        self.assertTrue(check_dexi_value(5))
        self.assertTrue(check_dexi_value(5.5))
        self.assertTrue(check_dexi_value(set()))
        self.assertTrue(check_dexi_value({1,2}))
        self.assertTrue(check_dexi_value([1,2]))
        self.assertTrue(check_dexi_value([]))
        self.assertFalse(check_dexi_value({1: 2}))
        self.assertFalse(check_dexi_value(lambda x: x))
        self.assertFalse(check_dexi_value("*"))
        self.assertFalse(check_dexi_value(""))
        self.assertFalse(check_dexi_value("Undefined"))
        self.assertFalse(check_dexi_value("defined"))
        self.assertFalse(check_dexi_value({1,"a"}))
        self.assertFalse(check_dexi_value((1,"a")))
        self.assertFalse(check_dexi_value([1,"a"]))
        self.assertFalse(check_dexi_value({"a": 2}))
        self.assertFalse(check_dexi_value({1: "x"}))
        self.assertFalse(check_dexi_value({1.1: 2}))

    def test_check_dexi_value_specification(self):
        self.assertTrue(check_dexi_value_specification(None))
        self.assertTrue(check_dexi_value_specification(5))
        self.assertTrue(check_dexi_value_specification(5.5))
        self.assertTrue(check_dexi_value_specification(set()))
        self.assertTrue(check_dexi_value_specification({1,2}))
        self.assertTrue(check_dexi_value_specification([1,2]))
        self.assertTrue(check_dexi_value_specification([]))
        self.assertTrue(check_dexi_value_specification({1: 2}))
        self.assertFalse(check_dexi_value_specification(lambda x: x))
        self.assertTrue(check_dexi_value_specification("*"))
        self.assertTrue(check_dexi_value_specification(""))
        self.assertTrue(check_dexi_value_specification("Undefined"))
        self.assertTrue(check_dexi_value_specification("defined"))
        self.assertTrue(check_dexi_value_specification({1,"a"}))
        self.assertTrue(check_dexi_value_specification((1,"a")))
        self.assertFalse(check_dexi_value_specification([1,"a"]))
        self.assertTrue(check_dexi_value_specification({"a": 2}))
        self.assertFalse(check_dexi_value_specification({1: "x"}))
        self.assertFalse(check_dexi_value_specification({1.1: 2}))

    def test_dexi_value_as_set(self):
        self.assertEqual(dexi_value_as_set(None), None)
        self.assertEqual(dexi_value_as_set("*"), None)
        self.assertEqual(dexi_value_as_set(""), None)
        self.assertEqual(dexi_value_as_set(3), {3})
        self.assertEqual(dexi_value_as_set({2, 3}), {2, 3})
        self.assertEqual(dexi_value_as_set((2, 3)), {2, 3})
        self.assertEqual(dexi_value_as_set((2, "a", 3)), None)
        self.assertEqual(dexi_value_as_set((2, 3.3, 3)), None)
        self.assertEqual(dexi_value_as_set([0.0, 1.0, 1.0, 0.5]), {1, 2, 3})
        self.assertEqual(dexi_value_as_set([0.0, 1.0, 1.0, 0.0], True), {1, 2})
        self.assertEqual(dexi_value_as_set([0.0, 1.0, 1.0, 0.5], True), None)
        self.assertEqual(dexi_value_as_set({1: 1, 2: 1}), {1, 2})
        self.assertEqual(dexi_value_as_set({1: 1, 2: 1, 3: 0.5}, strict = True), None)
        self.assertEqual(dexi_value_as_set({1: 1, 2: 1, 3: 0.5}), {1, 2, 3})

    def test_dexi_value_as_distr(self):
        self.assertEqual(dexi_value_as_distr(None), None)
        self.assertEqual(dexi_value_as_distr("*"), None)
        self.assertEqual(dexi_value_as_distr(""), None)
        self.assertEqual(dexi_value_as_distr(3), [0, 0, 0, 1])
        self.assertEqual(dexi_value_as_distr(-3), [])
        self.assertEqual(dexi_value_as_distr({2, 3}), [0, 0, 1, 1])
        self.assertEqual(dexi_value_as_distr((2, 3)), [0, 0, 1, 1])
        self.assertEqual(dexi_value_as_distr((2, "a", 3)), None)
        self.assertEqual(dexi_value_as_distr((2, 3.3, 3)), None)
        self.assertEqual(dexi_value_as_distr([0.0, 1.0, 1.0, 0.5]), [0.0, 1.0, 1.0, 0.5])
        self.assertEqual(dexi_value_as_distr({1:2, 2:3}), [0, 2, 3])

    def test_dexi_reduce_dexi_value(self):
        self.assertEqual(reduce_dexi_value(None), None)
        self.assertEqual(reduce_dexi_value(1), 1)
        self.assertEqual(reduce_dexi_value("*"), "*")
        self.assertEqual(reduce_dexi_value({2}), 2)
        self.assertEqual(reduce_dexi_value((2,)), 2)
        self.assertEqual(reduce_dexi_value((2,2)), 2)
        self.assertEqual(reduce_dexi_value((2,3)), {2,3})
        self.assertEqual(reduce_dexi_value([0, 1, 0]), 1)
        self.assertEqual(reduce_dexi_value([0, 0.5, 0]), [0, 0.5, 0])
        self.assertEqual(reduce_dexi_value([0, 1, 1, 0]), {1, 2})
        self.assertEqual(reduce_dexi_value({1:1, 2:1}), {1, 2})
        self.assertEqual(reduce_dexi_value({1:0.5, 2:1}), [0, 0.5, 1])

    def test_dexi_reduce_set(self):
        self.assertEqual(reduce_set(None), None)
        self.assertEqual(reduce_set(set()), None)
        self.assertEqual(reduce_set({}), {})
        self.assertEqual(reduce_set({2}), 2)
        self.assertEqual(reduce_set({1, 2}), {1, 2})

    def test_aggregate_value(self):
        self.assertEqual(aggregate_value(None), None)
        self.assertEqual(aggregate_value(4), 4)
        self.assertEqual(aggregate_value(4.3), 4.3)
        self.assertEqual(aggregate_value((), aggregate = "mean"), None)
        self.assertEqual(aggregate_value((5, 5), aggregate = "mean"), 5)
        self.assertEqual(aggregate_value((5, 5), aggregate = "min"), 5)
        self.assertEqual(aggregate_value((5, 5), aggregate = "max"), 5)
        self.assertEqual(aggregate_value((3, 5), aggregate = "mean"), 4)
        self.assertEqual(aggregate_value((3, 5), aggregate = "min"), 3)
        self.assertEqual(aggregate_value(set(), aggregate = "mean"), None)
        self.assertEqual(aggregate_value({}, aggregate = "mean"), None)
        self.assertEqual(aggregate_value((3, 5), aggregate = "max"), 5)
        self.assertEqual(aggregate_value({5}, aggregate = "mean"), 5)
        self.assertEqual(aggregate_value({5}, aggregate = "min"), 5)
        self.assertEqual(aggregate_value({5}, aggregate = "max"), 5)
        self.assertEqual(aggregate_value({3, 5}, aggregate = "mean"), 4)
        self.assertEqual(aggregate_value({3, 5}, aggregate = "min"), 3)
        self.assertEqual(aggregate_value({3, 5}, aggregate = "max"), 5)
        self.assertEqual(aggregate_value([], aggregate = "mean"), None)
        self.assertEqual(aggregate_value([0, 0.5, 0.25, 0.25], aggregate = "mean"), 0.5 + 2*0.25 + 3*0.25)
        self.assertEqual(aggregate_value([0, 0.5, 0.25, 0.25], aggregate = "min"), 1)
        self.assertEqual(aggregate_value([0, 0.5, 0.25, 0.25], aggregate = "max"), 3)
        self.assertEqual(aggregate_value([0, 0.5, 0.25, 0.25], aggregate = "mean", interpret = "set"), 2)
        self.assertEqual(aggregate_value([0, 0.5, 0.25, 0.25], aggregate = "min", interpret = "set"), 1)
        self.assertEqual(aggregate_value([0, 0.5, 0.25, 0.25], aggregate = "max", interpret = "set"), 3)
        with self.assertRaises(AssertionError):
            aggregate_value([0, 0.5, 0.25, 0.25], aggregate = "minmax")
        self.assertEqual(aggregate_value("str", aggregate = "min"), None)
        with self.assertRaises(Exception):
            aggregate_value([[0], "str", 0.25, 0.25], aggregate = "max")

    def test_compare_values(self):
        self.assertEqual(compare_values(None, 1), None)
        self.assertEqual(compare_values(1, 1), 0)
        self.assertEqual(compare_values(0, 1), -1)
        self.assertEqual(compare_values(1, 0), +1)
        self.assertEqual(compare_values(set(), set()), 0)
        self.assertEqual(compare_values({1}, set()), None)
        self.assertEqual(compare_values({1}, None), None)
        self.assertEqual(compare_values({1,2}, {1,2}), 0)
        self.assertEqual(compare_values({1,2}, {1,3}), None)
        self.assertEqual(compare_values({1,2}, {3,4}), -1)
        self.assertEqual(compare_values({1,2}, {2,4}), None)
        self.assertEqual(compare_values({1,2}, {2.1,4}), None)
        self.assertEqual(compare_values({1,2.05}, {2.1,4}), None)
        self.assertEqual(compare_values({3,4}, {3,4}), 0)
        self.assertEqual(compare_values({5}, {3,4}), +1)
        self.assertEqual(compare_values({5}, 2), +1)
        self.assertEqual(compare_values({5,2}, 2), None)
        self.assertEqual(compare_values({5,3}, 2), +1)
        self.assertEqual(compare_values([0,0,0,1,0,1], 2), +1)
        self.assertEqual(compare_values([0,0,0,1,0,1], 5), None)

    def test_compare_values_by_preference(self):
        self.assertEqual(compare_values_by_preference(None, 1, DexiOrder.ASCENDING), None)
        self.assertEqual(compare_values_by_preference(None, 1, DexiOrder.DESCENDING), None)
        self.assertEqual(compare_values_by_preference(None, 1, DexiOrder.NONE), None)
        self.assertEqual(compare_values_by_preference(1, 1, DexiOrder.ASCENDING), 0)
        self.assertEqual(compare_values_by_preference(1, 1, DexiOrder.DESCENDING), 0)
        self.assertEqual(compare_values_by_preference(1, 1, DexiOrder.NONE), 0)
        self.assertEqual(compare_values_by_preference(1, 2, DexiOrder.ASCENDING), -1)
        self.assertEqual(compare_values_by_preference(1, 2, DexiOrder.DESCENDING), +1)
        self.assertEqual(compare_values_by_preference(1, 2, DexiOrder.NONE), None)
        self.assertEqual(compare_values_by_preference(3, 2, DexiOrder.ASCENDING), +1)
        self.assertEqual(compare_values_by_preference(3, 2, DexiOrder.DESCENDING), -1)
        self.assertEqual(compare_values_by_preference(3, 2, DexiOrder.NONE), None)

    def test_DexiValues(self):
        v = DexiValues(None)
        self.assertEqual(v.value, None)
        self.assertEqual(v.value_type(), DexiValueType.NONE)
        self.assertTrue(v.check_value())
        self.assertEqual(v.as_set(), None)
        self.assertEqual(v.as_distr(), None)
        self.assertEqual(v.reduce(), None)

        v = DexiValues(1)
        self.assertEqual(v.value, 1)
        self.assertEqual(v.value_type(), DexiValueType.INT)
        self.assertTrue(v.check_value())
        self.assertEqual(v.as_set(), {1})
        self.assertEqual(v.as_distr(), [0, 1])
        self.assertEqual(v.reduce(), 1)

        v = DexiValues((1,))
        self.assertEqual(v.value, (1,))
        self.assertEqual(v.value_type(), DexiValueType.TUPLE)
        self.assertFalse(v.check_value())
        self.assertEqual(v.as_set(), {1})
        self.assertEqual(v.as_distr(), [0, 1])
        self.assertEqual(v.reduce(), 1)
        v.reduce_value()
        self.assertEqual(v.value, 1)
        self.assertEqual(v.value_type(), DexiValueType.INT)
        self.assertEqual(v.reduce(), 1)

        v = DexiValues([0, 1])
        self.assertEqual(v.value, [0, 1])
        self.assertEqual(v.value_type(), DexiValueType.LIST)
        self.assertTrue(v.check_value())
        self.assertEqual(v.as_set(), {1})
        self.assertEqual(v.as_distr(), [0, 1])
        self.assertEqual(v.reduce(), 1)
        v.reduce_value()
        self.assertEqual(v.value, 1)
        self.assertEqual(v.value_type(), DexiValueType.INT)
        self.assertEqual(v.reduce(), 1)

if __name__ == '__main__':
    unittest.main()
